#include <RaptorQ/v1/Octet.hpp>
#include <RaptorQ/v1/Matrix.hpp>
#include <RaptorQ/v1/util/Matrix_Row.hpp>
#include <memory>
#include <utility>
#include <iostream>
#include <chrono>
#include <cstdlib>
#include <stdlib.h>
#include <ctime>
#include "catch2/catch.hpp"

using namespace RaptorQ__v1::Impl;

TEST_CASE( "Test eigen compatibility", "[Eigen compatibility]" )
{
    std::time_t result = std::time(nullptr);
    std::cout << "Random seed used in [Eigen compatibility]: " << (int)result << std::endl;
    srand( result );

    uint8_t number;
    DenseMtx original = DenseMtx(10, 1316);
    for ( int i = 0; i < original.rows(); i ++ )  {
        for ( int j = 0; j < original.cols(); j ++ ) {
            original(i,j) = static_cast<uint8_t> (rand() % 256);
        }
    }

    // Test for all types (Will use best available as fallback if not available)
    const Matrix_Row::SIMD simd_types[] = { Matrix_Row::SIMD::NONE, Matrix_Row::SIMD::SSSE3, Matrix_Row::SIMD::AVX2, Matrix_Row::SIMD::AUTO };
    for (int type = 0;type < 3;type++) {
        // Set SIMD type: in production use "Matrix_Row::SIMD::AUTO"
        Matrix_Row::get_instance().init_simd(simd_types[type]);
        // Test all numbers
        for (int i = 0;i < 256;i++) {
            uint8_t number = i;
            DenseMtx A = DenseMtx(original);
            Matrix B = Matrix(original);
            REQUIRE( A == B.to_eigen() );

            // row_multiply_add
            A.row (0) += A.row (1) * Octet(number);
            B.row_multiply_add(0, B, 1, number);
            REQUIRE( A == B.to_eigen() );

            // row_multiply_sub
            A.row (1) -= A.row (2) * Octet(number);
            B.row_multiply_sub(1, B, 2, number);
            REQUIRE( A == B.to_eigen() );

            // row_div
            A.row (3) /= Octet(number);
            B.row_div(3, number);
            REQUIRE( A == B.to_eigen() );

            // row_add
            A.row (4) += A.row (5);
            B.row_add(4, B, 5);
            REQUIRE( A == B.to_eigen() );

            // row_swap
            A.row (4).swap (A.row (5));
            B.row_swap(4, 5);
            A.row (9).swap (A.row (3));
            B.row_swap(9, 3);
            A.row (2).swap (A.row (0));
            B.row_swap(2, 0);
            REQUIRE( A == B.to_eigen() );

            // col_swap
            A.col (4).swap (A.col (5));
            B.col_swap(4, 5);
            A.col (78).swap (A.col (1311));
            B.col_swap(78, 1311);
            A.col (0).swap (A.col (13));
            B.col_swap(0, 13);
            REQUIRE( A == B.to_eigen() );

            // row_assign
            A.row (5) = A.row (6);
            B.row_assign(5, B, 6);
            REQUIRE( A == B.to_eigen() );
        }
    }
}
