#include <RaptorQ/v1/Octet.hpp>
#include <RaptorQ/v1/Matrix.hpp>
#include <RaptorQ/v1/util/Matrix_Row.hpp>
#include "RaptorQ/v1/Shared_Computation/Decaying_LF.hpp"
#include <memory>
#include <utility>
#include <iostream>
#include <chrono>
#include <cstdlib>
#include "catch2/catch.hpp"

using namespace RaptorQ__v1::Impl;

TEST_CASE( "Store all zeroes", "[Matrix store]" )
{
    Matrix A = Matrix(4, 4);
    std::vector<uint8_t> raw = mtx_to_raw(A);
    Matrix B = raw_to_mtx(raw, A.rows(), A.cols());

    REQUIRE( raw.size() == 0 );
    REQUIRE( A == B );

    A = Matrix(255, 255);
    raw = mtx_to_raw(A);
    B = raw_to_mtx(raw, A.rows(), A.cols());

    REQUIRE( raw.size() == 0 );
    REQUIRE( A == B );

    A = Matrix(256, 256);
    raw = mtx_to_raw(A);
    B = raw_to_mtx(raw, A.rows(), A.cols());

    REQUIRE( raw.size() == 0 );
    REQUIRE( A == B );

    A = Matrix(6000, 6000);
    raw = mtx_to_raw(A);
    B = raw_to_mtx(raw, A.rows(), A.cols());

    REQUIRE( raw.size() == 0 );
    REQUIRE( A == B );
}

TEST_CASE( "Single value", "[Matrix store]" )
{
    Matrix A = Matrix(4, 6);
    A(1, 2) = 34;
    std::vector<uint8_t> raw = mtx_to_raw(A);
    Matrix B = raw_to_mtx(raw, A.rows(), A.cols());

    REQUIRE( raw.size() == 3 );
    REQUIRE( A == B );

    A = Matrix(255, 257);
    A(1, 2) = 34;
    raw = mtx_to_raw(A);
    B = raw_to_mtx(raw, A.rows(), A.cols());

    REQUIRE( raw.size() == 5 );
    REQUIRE( A == B );

    A = Matrix(256, 258);
    A(1, 2) = 34;
    raw = mtx_to_raw(A);
    B = raw_to_mtx(raw, A.rows(), A.cols());

    REQUIRE( raw.size() == 5 );
    REQUIRE( A == B );

    A = Matrix(6000, 6002);
    A(1, 2) = 34;
    raw = mtx_to_raw(A);
    B = raw_to_mtx(raw, A.rows(), A.cols());

    REQUIRE( raw.size() == 5 );
    REQUIRE( A == B );
}

TEST_CASE( "Store 4x4 matrix", "[Matrix store]" )
{
    Matrix A = Matrix(4, 4);
    for (int i = 0; i < 4;i++) {
        for (int j = 0; j < 4;j++) {
            A(i, j) = i * 4 + j;
        }
    }

    std::vector<uint8_t> raw = mtx_to_raw(A);
    Matrix B = raw_to_mtx(raw, 4, 4);

    REQUIRE( raw.size() == 3 * 16 - 3 );
    REQUIRE( A == B );

    A(0, 0) = 54;
    raw = mtx_to_raw(A);
    REQUIRE( raw.size() == 3 * 16 );

    A(0, 0) = 0;
    A(1, 1) = 0;
    A(1, 2) = 0;
    A(2, 0) = 0;
    A(2, 1) = 0;
    A(3, 3) = 0;
    raw = mtx_to_raw(A);
    REQUIRE( raw.size() == 3 * 10 );
    raw = mtx_to_raw(A);
    B = raw_to_mtx(raw, 4, 4);
    REQUIRE( A == B );

    A(0, 1) = 0;
    A(0, 2) = 0;
    A(0, 3) = 0;
    A(1, 0) = 0;
    A(1, 3) = 0;
    A(2, 2) = 0;
    A(2, 3) = 0;
    A(3, 0) = 0;
    raw = mtx_to_raw(A);
    REQUIRE( raw.size() == 3 * 2 );
    raw = mtx_to_raw(A);
    B = raw_to_mtx(raw, 4, 4);
    REQUIRE( A == B );
}
