#include <RaptorQ/v1/Octet.hpp>
#include <RaptorQ/v1/Matrix.hpp>
#include <memory>
#include <utility>
#include <iostream>
#include <chrono>
#include <cstdlib>
#include "catch2/catch.hpp"

using namespace RaptorQ__v1::Impl;

TEST_CASE( "Row multiply add", "[Dense Matrix]" )
{
    const Matrix_Row::SIMD simd_types[] = { Matrix_Row::SIMD::NONE, Matrix_Row::SIMD::SSSE3, Matrix_Row::SIMD::AVX2, Matrix_Row::SIMD::AUTO };
    for (Matrix_Row::SIMD type : simd_types) {
        // Set SIMD type: in production use "Matrix_Row::SIMD::AUTO"
        Matrix_Row::get_instance().init_simd(type);
        Matrix A = Matrix(2, 4);

        A(0, 0) = 5;
        A(0, 1) = 0;
        A(0, 2) = 1;
        A(0, 3) = 117;
        A.row_multiply_add(1, A, 0, 2);

        REQUIRE( A(1, 0) ==  10 );
        REQUIRE( A(1, 1) ==   0 );
        REQUIRE( A(1, 2) ==   2 );
        REQUIRE( A(1, 3) == 234 );

        A.row_multiply_add(0, A, 0, 2);

        REQUIRE( A(0, 0) == (5 ^ 10) );
        REQUIRE( A(0, 1) == (0 ^ 0) );
        REQUIRE( A(0, 2) == (1 ^ 2) );
        REQUIRE( A(0, 3) == (117 ^ 234) );
    }
}

TEST_CASE( "Row multiply sub", "[Dense Matrix]" )
{
    const Matrix_Row::SIMD simd_types[] = { Matrix_Row::SIMD::NONE, Matrix_Row::SIMD::SSSE3, Matrix_Row::SIMD::AVX2, Matrix_Row::SIMD::AUTO };
    for (Matrix_Row::SIMD type : simd_types) {
        // Set SIMD type: in production use "Matrix_Row::SIMD::AUTO"
        Matrix_Row::get_instance().init_simd(type);
        Matrix A = Matrix(2, 4);

        A(0, 0) = 5;
        A(0, 1) = 0;
        A(0, 2) = 1;
        A(0, 3) = 117;
        A.row_multiply_sub(1, A, 0, 2);

        REQUIRE( A(1, 0) ==  10 );
        REQUIRE( A(1, 1) ==   0 );
        REQUIRE( A(1, 2) ==   2 );
        REQUIRE( A(1, 3) == 234 );

        A.row_multiply_sub(0, A, 0, 2);

        REQUIRE( A(0, 0) == (5 ^ 10) );
        REQUIRE( A(0, 1) == (0 ^ 0) );
        REQUIRE( A(0, 2) == (1 ^ 2) );
        REQUIRE( A(0, 3) == (117 ^ 234) );
    }
}

TEST_CASE( "Row div", "[Dense Matrix]" )
{
    const Matrix_Row::SIMD simd_types[] = { Matrix_Row::SIMD::NONE, Matrix_Row::SIMD::SSSE3, Matrix_Row::SIMD::AVX2, Matrix_Row::SIMD::AUTO };
    for (Matrix_Row::SIMD type : simd_types) {
        // Set SIMD type: in production use "Matrix_Row::SIMD::AUTO"
        Matrix_Row::get_instance().init_simd(type);
        Matrix A = Matrix(1, 4);

        A(0, 0) = 5;
        A(0, 1) = 0;
        A(0, 2) = 1;
        A(0, 3) = 117;
        A.row_div(0, 0);

        REQUIRE( A(0, 0) ==   5 );
        REQUIRE( A(0, 1) ==   0 );
        REQUIRE( A(0, 2) ==   1 );
        REQUIRE( A(0, 3) == 117 );

        A.row_div(0, 37);

        REQUIRE( A(0, 0) == 19 ); //TODO: verify that this is correct
        REQUIRE( A(0, 1) ==  0 );
        REQUIRE( A(0, 2) == 86 ); //TODO: verify that this is correct
        REQUIRE( A(0, 3) == 44 ); //TODO: verify that this is correct
    }
}

TEST_CASE( "Row swap", "[Dense Matrix]" )
{
    Matrix A = Matrix(2, 31);

    A(0, 0)  = 5;
    A(0, 1)  = 0;
    A(0, 2)  = 1;
    A(0, 3)  = 117;
    A(0, 30) = 7;

    A(1, 0)  = 0;
    A(1, 1)  = 3;
    A(1, 2)  = 15;
    A(1, 3)  = 23;
    A(1, 29) = 21;

    A.row_swap(0, 1);

    REQUIRE( A(0, 0)  ==  0 );
    REQUIRE( A(0, 1)  ==  3 );
    REQUIRE( A(0, 2)  == 15 );
    REQUIRE( A(0, 3)  == 23 );
    REQUIRE( A(0, 29) == 21 );

    REQUIRE( A(1, 0)  ==   5 );
    REQUIRE( A(1, 1)  ==   0 );
    REQUIRE( A(1, 2)  ==   1 );
    REQUIRE( A(1, 3)  == 117 );
    REQUIRE( A(1, 30) ==   7 );
}

TEST_CASE( "Col swap", "[Dense Matrix]" )
{
    Matrix A = Matrix(2, 31);

    A(0, 0)  = 5;
    A(0, 1)  = 0;
    A(0, 2)  = 1;
    A(0, 3)  = 117;
    A(0, 30) = 7;

    A(1, 0)  = 0;
    A(1, 1)  = 3;
    A(1, 2)  = 15;
    A(1, 3)  = 23;
    A(1, 29) = 21;

    A.col_swap(0, 1);

    REQUIRE( A(0, 0)  ==  0 );
    REQUIRE( A(1, 0)  ==  3 );
    REQUIRE( A(0, 1)  ==  5 );
    REQUIRE( A(1, 1)  ==  0 );

    A.col_swap(3, 7);

    REQUIRE( A(0, 3)  ==   0 );
    REQUIRE( A(1, 3)  ==   0 );
    REQUIRE( A(0, 7)  == 117 );
    REQUIRE( A(1, 7)  ==  23 );
}

TEST_CASE( "Row add", "[Dense Matrix]" )
{
    const Matrix_Row::SIMD simd_types[] = { Matrix_Row::SIMD::NONE, Matrix_Row::SIMD::SSSE3, Matrix_Row::SIMD::AVX2, Matrix_Row::SIMD::AUTO };
    for (Matrix_Row::SIMD type : simd_types) {
        Matrix A = Matrix(2, 31);

        A(0, 0)  = 5;
        A(0, 1)  = 0;
        A(0, 2)  = 1;
        A(0, 3)  = 117;
        A(0, 30) = 7;

        A(1, 0)  = 0;
        A(1, 1)  = 3;
        A(1, 2)  = 15;
        A(1, 3)  = 23;
        A(1, 29) = 21;

        A.row_add(0, A, 1);

        REQUIRE( A(0, 0)  == (5 ^ 0) );
        REQUIRE( A(0, 1)  == (0 ^ 3) );
        REQUIRE( A(0, 2)  == (1 ^ 15) );
        REQUIRE( A(0, 3)  == (117 ^ 23) );
        REQUIRE( A(0, 29) == (0 ^ 21) );
        REQUIRE( A(0, 30) == (7 ^ 0) );
    }
}

TEST_CASE( "Row count zeroes", "[Dense Matrix]" )
{
    const Matrix_Row::SIMD simd_types[] = { Matrix_Row::SIMD::NONE, Matrix_Row::SIMD::SSSE3, Matrix_Row::SIMD::AVX2, Matrix_Row::SIMD::AUTO };
    for (Matrix_Row::SIMD type : simd_types) {
        // Set SIMD type: in production use "Matrix_Row::SIMD::AUTO"
        Matrix_Row::get_instance().init_simd(type);

        Matrix A = Matrix(3, 37);

        A(0, 0)  = 1;
        A(0, 36) = 7;

        REQUIRE( A.non_zeros_in_row(0, 0, 37) == 2 );
        REQUIRE( A.non_zeros_in_row(0, 1, 36) == 0 );

        A(1, 0)  = 1;
        A(1, 5)  = 255;
        A(1, 36) = 7;

        REQUIRE( A.non_zeros_in_row(1, 0, 37) == 3 );
        REQUIRE( A.non_zeros_in_row(1, 1, 36) == 1 );

        Matrix B = Matrix(2, 512);
        for (int i = 0; i < 512; i++) {
            B(0, i) = 1;
            B(1, i) = (i % 2);
        }
        REQUIRE( B.non_zeros_in_row(0, 0, 512) == 512 );
        REQUIRE( B.non_zeros_in_row(1, 0, 512) == 256 );
        REQUIRE( B.non_zeros_in_row(2, 0, 512) == 0 );
        REQUIRE( B.non_zeros_in_row(0, 6, 506) == 500 );
        REQUIRE( B.non_zeros_in_row(1, 6, 506) == 250 );
        REQUIRE( B.non_zeros_in_row(2, 6, 506) == 0 );
        REQUIRE( B.non_zeros_in_row(0, 6, 6) == 0 );
        REQUIRE( B.non_zeros_in_row(0, 6, 7) == 1 );
        REQUIRE( B.non_zeros_in_row(0, 31, 32) == 1 );
        REQUIRE( B.non_zeros_in_row(0, 32, 33) == 1 );
        REQUIRE( B.non_zeros_in_row(0, 31, 49) == 18 );
        REQUIRE( B.non_zeros_in_row(2, 6, 7) == 0 );
        REQUIRE( B.non_zeros_in_row(2, 31, 32) == 0 );
        REQUIRE( B.non_zeros_in_row(2, 32, 33) == 0 );
        REQUIRE( B.non_zeros_in_row(2, 31, 49) == 0 );

        Matrix C = Matrix(2, 65539);
        for (int i = 0; i < 65539; i++) {
            C(0, i) = 1;
            C(1, i) = (i % 2);
        }
        REQUIRE( C.non_zeros_in_row(0, 0, 65539) == 65539 );
        REQUIRE( C.non_zeros_in_row(1, 0, 65539) == 32769 );

        Matrix D = Matrix(2, 1748000);
        for (int i = 0; i < 1748000; i++) {
            D(0, i) = 1;
            D(1, i) = (i % 2);
        }
        REQUIRE( D.non_zeros_in_row(0, 0, 1748000) == 1748000 );
        REQUIRE( D.non_zeros_in_row(1, 0, 1748000) == 874000 );
    }
}

TEST_CASE( "Row first zero", "[Dense Matrix]" )
{
    const Matrix_Row::SIMD simd_types[] = { Matrix_Row::SIMD::NONE, Matrix_Row::SIMD::SSSE3, Matrix_Row::SIMD::AVX2, Matrix_Row::SIMD::AUTO };
    for (Matrix_Row::SIMD type : simd_types) {
        // Set SIMD type: in production use "Matrix_Row::SIMD::AUTO"
        Matrix_Row::get_instance().init_simd(type);

        Matrix A = Matrix(3, 37);

        A(0, 0)  = 1;
        A(0, 36) = 7;

        REQUIRE( A.first_non_zero_in_row(0, 0, 37) == 0 );
        REQUIRE( A.first_non_zero_in_row(0, 1, 37) == 36 );

        A(1, 1)  = 1;
        A(1, 5)  = 255;

        REQUIRE( A.first_non_zero_in_row(1, 0, 37) == 1 );
        REQUIRE( A.first_non_zero_in_row(1, 2, 37) == 5 );

        Matrix B = Matrix(1, 512);
        for (int i = 0; i < 512; i++) {
            B(0, i) = 1;
            B(1, i) = (i % 2);
        }
        for (int i = 0; i < 512; i++) {
            REQUIRE( B.first_non_zero_in_row(0, i, 512) == i );
            REQUIRE( B.first_non_zero_in_row(1, i, 512) == i + ((i % 2) ? 0 : 1) );
        }

        Matrix C = Matrix(2, 65539);
        C(0, 65537) = 1;
        C(0, 65538) = 1;
        C(1, 32769) = 1;
        C(1, 65538) = 1;
        REQUIRE( C.first_non_zero_in_row(0, 0, 65539)     == 65537 );
        REQUIRE( C.first_non_zero_in_row(0, 65538, 65539) == 65538 );
        REQUIRE( C.first_non_zero_in_row(1, 0, 65539)     == 32769 );
        REQUIRE( C.first_non_zero_in_row(1, 32770, 65539) == 65538 );
    }
}
